<?php
// $Id: operations.csv.inc,v 1.2 2009/03/24 19:46:37 jpetso Exp $
/**
 * @file
 * CSV Transformations -
 * Transformations for processing CSV (Comma Separated Values) data.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */

// Also see http://www.rfc-editor.org/rfc/rfc4180.txt which describes the
// standardized version of the CSV format. PHP's fgetcsv() and str_getcsv()
// functions do not quite conform to that standard because they use an
// escape character ('\') instead of using a doubled double-quote ('""') for
// double-quote escapings inside a quoted field.
//
// Despite the massive amount of example CSV parsers in the PHP documentation's
// user comment section, none seem to handle this standard with all its traits
// (quote escaping, quoted fields spanning multiple lines, etc.) correctly
// and in a readable way. So here's another custom implementation. (Bah.)


/**
 * Helper class for providing implementations for standard input/output info.
 */
class TfCSVConnectionInfoHelper {
  public static function csvTextInfo($propertyKey) {
    switch ($propertyKey) {
      case 'label':
        return t('CSV text');

      case 'expectedSchema':
        return new TfWritableSchemaElement('csvText', 'php:type:string');
    }
  }

  public static function delimiterInfo($propertyKey) {
    switch ($propertyKey) {
      case 'required':
        return FALSE;

      case 'label':
        return t('Delimiter');

      case 'defaultValue':
        return ',';

      case 'expectedSchema':
        return new TfWritableSchemaElement('delimiter', 'php:type:string');
    }
  }

  public static function skipFirstLineInfo($propertyKey) {
    switch ($propertyKey) {
      case 'required':
        return FALSE;

      case 'label':
        return t('Skip first line');

      case 'defaultValue':
        return FALSE;

      case 'expectedSchema':
        return new TfWritableSchemaElement('skipFirstLine', 'php:type:boolean');
    }
  }

  public static function columnNamesInfo($propertyKey) {
    switch ($propertyKey) {
      case 'required':
        return FALSE;

      case 'label':
        return t('Column names');

      case 'defaultValue':
        return array();

      case 'expectedSchema':
        $columnNameElement = new TfWritableSchemaElement('name', 'php:type:string');
        $columnNameElement->setLabelRequirements(TfSchemaElement::NoMatchRequired);
        $cardinality = new TfWritableSchemaCardinality($columnNameElement,
          1, TfSchemaCardinality::Unbounded
        );
        $columnNamesElement = new TfWritableSchemaElement('columnNames');
        $columnNamesElement->setChildCollection(new TfSchemaSequence());
        $columnNamesElement->childCollection()->addMember($cardinality);
        return $columnNamesElement;
    }
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_csv_operation_TfRecordsFromCSVText() {
  return array(
    'label' => t('CSV text to list of records'),
    'description' => t('Transforms plaintext CSV (comma separated values) data into a list of records, each record consisting of the fields in the respective text line.'),
  );
}

class TfRecordsFromCSVText extends TfOperation {
  private $lineTransformation;

  /**
   * Overriding TfOperation::initialize().
   */
  public function initialize(Array $typeOptions = array()) {
    $this->lineTransformation = TfOperation::load('TfRecordsFromCSVTextLines');
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('csvText', 'delimiter', 'skipFirstLine', 'columnNames');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'csvText') {
      return TfCSVConnectionInfoHelper::csvTextInfo($propertyKey);
    }
    elseif ($inputKey == 'delimiter') {
      return TfCSVConnectionInfoHelper::delimiterInfo($propertyKey);
    }
    elseif ($inputKey == 'skipFirstLine') {
      return TfCSVConnectionInfoHelper::skipFirstLineInfo($propertyKey);
    }
    elseif ($inputKey == 'columnNames') {
      return TfCSVConnectionInfoHelper::columnNamesInfo($propertyKey);
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return $this->lineTransformation->outputs();
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    return $this->lineTransformation->outputInfo($outputKey, $propertyKey);
  }

  /**
   * Overriding TfOperation::inputChanged().
   */
  protected function inputChanged($key) {
    if ($key != 'csvText') {
      $this->lineTransformation->setInput($key, $this->input($key));
    }
    parent::inputChanged($key);
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $lines = explode("\n", $this->input('csvText')->data());
    $this->lineTransformation->setInput('csvTextLines', $lines);
    $lineOutput = $this->lineTransformation->output();

    if (!$lineOutput->isValid()) {
      $output->setErrorMessage($lineOutput->errorMessage());
      return;
    }
    foreach ($lineOutput as $key => $data) {
      $output->set($key, $data);
    }
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_csv_operation_TfRecordsFromCSVTextLines() {
  return array(
    'label' => t('CSV text lines to list of records'),
    'description' => t('Transforms a list of text lines in CSV (comma separated values) format into a list of records, each record consisting of the fields in the respective text line.'),
  );
}

class TfRecordsFromCSVTextLines extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('csvTextLines', 'delimiter', 'skipFirstLine', 'columnNames');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'csvTextLines') {
      switch ($propertyKey) {
        case 'label':
          return t('CSV text lines');

        case 'expectedSchema':
          $lineElement = new TfWritableSchemaElement('line', 'php:type:string');
          $lineElement->setLabelRequirements(TfSchemaElement::NoMatchRequired);
          $cardinality = new TfWritableSchemaCardinality($lineElement,
            0, TfSchemaCardinality::Unbounded
          );
          $linesElement = new TfWritableSchemaElement('csvTextLines');
          $linesElement->setChildCollection(new TfSchemaSequence());
          $linesElement->childCollection()->addMember($cardinality);
          return $linesElement;
      }
    }
    elseif ($inputKey == 'delimiter') {
      return TfCSVConnectionInfoHelper::delimiterInfo($propertyKey);
    }
    elseif ($inputKey == 'skipFirstLine') {
      return TfCSVConnectionInfoHelper::skipFirstLineInfo($propertyKey);
    }
    elseif ($inputKey == 'columnNames') {
      return TfCSVConnectionInfoHelper::columnNamesInfo($propertyKey);
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('records');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'records') {
      switch ($propertyKey) {
        case 'label':
          return t('List of records');

        case 'expectedSchema':
          $recordsElement = new TfWritableSchemaElement('records');
          $recordsElement->setChildCollection(new TfSchemaSequence());

          foreach ($this->input('columnNames')->children() as $columnName) {
            $columnElement = new TfWritableSchemaElement($columnName, 'php:type:string');
            $recordsElement->childCollection()->addMember($columnElement);
          }

          if ($recordsElement->childCollection()->isEmpty()) { // no column names given
            $columnElement = new TfWritableSchemaElement('field', 'php:type:string');
            $cardinality = new TfWritableSchemaCardinality($columnElement,
              0, TfSchemaCardinality::Unbounded
            );
            $recordsElement->childCollection()->addMember($cardinality);
          }
          return $recordsElement;
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $columnNames = array();

    foreach ($this->input('columnNames')->children() as $columnName) {
      $columnNames[] = $columnName;
    }

    module_load_include('php', 'transformations_csv', 'CSVParser');

    $parser = new CSVParser();
    $parser->setDelimiter($this->input('delimiter')->data());
    $parser->setSkipFirstLine($this->input('skipFirstLine')->data());
    $parser->setColumnNames(empty($columnNames) ? FALSE : $columnNames);
    $rows = $parser->parse($this->input('csvTextLines')->children());

    $output->set('records', $rows);
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_csv_operation_TfCSVTextFromRecords() {
  return array(
    'label' => t('List of records to CSV text'),
    'description' => t('Transforms a list of records - each record consisting of a fixed number of string fields - into plaintext CSV (comma separated values) data.'),
  );
}

class TfCSVTextFromRecords extends TfOperation {
  private $lineTransformation;

  /**
   * Overriding TfOperation::initialize().
   */
  public function initialize(Array $typeOptions = array()) {
    $this->lineTransformation = TfOperation::load('TfCSVTextLineFromRecord');
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('records', 'delimiter', 'skipFirstLine', 'columnNames');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'records') {
      switch ($propertyKey) {
        case 'label':
          return t('List of records');

        case 'expectedSchema':
          $recordElement = $this->lineTransformation->inputProperty(
            'record', 'expectedSchema'
          );
          $cardinality = new TfWritableSchemaCardinality($recordElement,
            1, TfSchemaCardinality::Unbounded
          );
          $recordsElement = new TfWritableSchemaElement('records');
          $recordsElement->setChildCollection(new TfSchemaSequence());
          $recordsElement->childCollection()->addMember($cardinality);
          return $recordsElement;
      }
    }
    elseif ($inputKey == 'delimiter') {
      return TfCSVConnectionInfoHelper::delimiterInfo($propertyKey);
    }
    elseif ($inputKey == 'skipFirstLine') {
      return TfCSVConnectionInfoHelper::skipFirstLineInfo($propertyKey);
    }
    elseif ($inputKey == 'columnNames') {
      return TfCSVConnectionInfoHelper::columnNamesInfo($propertyKey);
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('csvText');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'csvText') {
      return TfCSVConnectionInfoHelper::csvTextInfo($propertyKey);
    }
  }

  /**
   * Overriding TfOperation::inputChanged().
   */
  protected function inputChanged($key) {
    if ($key == 'delimiter') {
      $this->lineTransformation->setInput($key, $this->input($key));
    }
    parent::inputChanged($key);
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $rows = array();
    $csvText = '';

    if ($this->input('skipFirstLine')->data() == TRUE) {
      // Insert the row names as additional line at the start of the array.
      $this->lineTransformation->setInput('record', $this->input('columnNames'));
      $lineOutput = $this->lineTransformation->output();

      if (!$lineOutput->isValid()) {
        $output->setErrorMessage($lineOutput->errorMessage());
        return;
      }
      $csvText .= $lineOutput->at('csvText') . "\n";
    }
    foreach ($this->input('records')->children() as $row) {
      $this->lineTransformation->setInput('record', $row);
      $lineOutput = $this->lineTransformation->output();

      if (!$lineOutput->isValid()) {
        $output->setErrorMessage($lineOutput->errorMessage());
        return;
      }
      $csvText .= $lineOutput->at('csvText')->data() . "\n";
    }
    $output->set('csvText', $csvText);
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_csv_operation_TfCSVTextLineFromRecord() {
  return array(
    'label' => t('List of strings to CSV text line'),
    'description' => t('Transforms a single record - consisting of a number of string fields - into plaintext CSV (comma separated values) data. The resulting text is a single line in most cases, except if a field contained a line break by itself.'),
  );
}

/**
 * A transformation to get plaintext CSV data for a single row (or line)
 * out of a string array. Not trailed by line breaks.
 */
class TfCSVTextLineFromRecord extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('record', 'delimiter');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'record') {
      switch ($propertyKey) {
        case 'label':
          return t('Record (list of strings)');

        case 'expectedSchema':
          $fieldElement = new TfWritableSchemaElement('field', 'php:type:string');
          $fieldElement->setLabelRequirements(TfSchemaElement::NoMatchRequired);
          $cardinality = new TfWritableSchemaCardinality($fieldElement,
            1, TfSchemaCardinality::Unbounded
          );
          $recordElement = new TfWritableSchemaElement('record');
          $recordElement->setChildCollection(new TfSchemaSequence());
          $recordElement->childCollection()->addMember($cardinality);
          return $recordElement;
      }
    }
    elseif ($inputKey == 'delimiter') {
      return TfCSVConnectionInfoHelper::delimiterInfo($propertyKey);
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('csvText');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'csvText') {
      return TfCSVConnectionInfoHelper::csvTextInfo($propertyKey);
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $fields = array();

    foreach ($this->input('record')->children() as $key => $field) {
      // Stick everything into double quotes and escape the real double quotes
      // by duplicating them.
      $fields[$key] = '"' . strtr($field, array('"' => '""')) . '"';
    }
    $output->set('csvText', implode($this->input('delimiter')->data(), $fields));
  }
}
